@@ -5,6 +5,8 @@
 #include "query.h"
 #include "request.h"
 #include "response.h"
+#include "tracing.h"
+#include "translate.h"
 #include "vfs.h"
 
 void gateway__init(struct gateway *g,
@@ -12,6 +14,7 @@ void gateway__init(struct gateway *g,
 		   struct registry *registry,
 		   struct raft *raft)
 {
+        tracef("gateway init");
 	g->config = config;
 	g->registry = registry;
 	g->raft = raft;
@@ -23,22 +26,51 @@ void gateway__init(struct gateway *g,
 	g->sql = NULL;
 	stmt__registry_init(&g->stmts);
 	g->barrier.data = g;
+	g->barrier.cb = NULL;
+	g->barrier.leader = NULL;
 	g->protocol = DQLITE_PROTOCOL_VERSION;
 }
 
 void gateway__close(struct gateway *g)
 {
-	stmt__registry_close(&g->stmts);
-	if (g->leader != NULL) {
-		if (g->stmt != NULL) {
+	tracef("gateway close");
+	if (g->leader == NULL) {
+		stmt__registry_close(&g->stmts);
+		return;
+	}
+
+	if (g->stmt != NULL) {
+		if (g->leader->inflight != NULL) {
+			tracef("finish inflight apply request");
 			struct raft_apply *req = &g->leader->inflight->req;
 			req->cb(req, RAFT_SHUTDOWN, NULL);
 			assert(g->req == NULL);
 			assert(g->stmt == NULL);
+		} else if (g->barrier.cb != NULL) {
+			tracef("finish inflight query barrier");
+			/* This is not a typo, g->barrier.req.cb is a wrapper
+			 * around g->barrier.cb and when called, will set g->barrier.cb to NULL.
+			 * */
+			struct raft_barrier *b = &g->barrier.req;
+			b->cb(b, RAFT_SHUTDOWN);
+			assert(g->barrier.cb == NULL);
+		} else if (g->leader->exec != NULL && g->leader->exec->barrier.cb != NULL) {
+			tracef("finish inflight exec barrier");
+			struct raft_barrier *b = &g->leader->exec->barrier.req;
+			b->cb(b, RAFT_SHUTDOWN);
+			assert(g->leader->exec == NULL);
+		} else if (g->req != NULL && g->req->type != DQLITE_REQUEST_QUERY
+					  && g->req->type != DQLITE_REQUEST_EXEC) {
+			/* Regular exec and query stmt's will be closed when the
+			 * registry is closed below. */
+			tracef("finalize exec_sql or query_sql type:%d", g->req->type);
+			sqlite3_finalize(g->stmt);
+			g->stmt = NULL;
 		}
-		leader__close(g->leader);
-		sqlite3_free(g->leader);
 	}
+	stmt__registry_close(&g->stmts);
+	leader__close(g->leader);
+	sqlite3_free(g->leader);
 }
 
 /* Declare a request struct and a response struct of the appropriate types and
@@ -92,6 +124,27 @@ void gateway__close(struct gateway *g)
 		return 0;                                    \
 	}
 
+#define FAIL_IF_CHECKPOINTING                                                  \
+	{                                                                      \
+		struct sqlite3_file *_file;                                    \
+		int _rv;                                                       \
+		_rv = sqlite3_file_control(g->leader->conn, "main",            \
+					   SQLITE_FCNTL_FILE_POINTER, &_file); \
+		assert(_rv == SQLITE_OK); /* Should never fail */              \
+                                                                               \
+		_rv = _file->pMethods->xShmLock(                               \
+		    _file, 1 /* checkpoint lock */, 1,                         \
+		    SQLITE_SHM_LOCK | SQLITE_SHM_EXCLUSIVE);                   \
+		if (_rv != 0) {                                                \
+			assert(_rv == SQLITE_BUSY);                            \
+			failure(req, SQLITE_BUSY, "checkpoint in progress");   \
+			return 0;                                              \
+		}                                                              \
+		_file->pMethods->xShmLock(                                     \
+		    _file, 1 /* checkpoint lock */, 1,                         \
+		    SQLITE_SHM_UNLOCK | SQLITE_SHM_EXCLUSIVE);                 \
+	}
+
 /* Encode fa failure response and invoke the request callback */
 static void failure(struct handle *req, int code, const char *message)
 {
@@ -112,6 +165,7 @@ static void failure(struct handle *req, int code, const char *message)
 
 static int handle_leader_legacy(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle leader legacy");
 	START(leader, server_legacy);
 	raft_id id;
 	raft_leader(req->gateway->raft, &id, &response.address);
@@ -124,6 +178,7 @@ static int handle_leader_legacy(struct handle *req, struct cursor *cursor)
 
 static int handle_leader(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle leader");
 	struct gateway *g = req->gateway;
 	raft_id id = 0;
 	const char *address = NULL;
@@ -137,6 +192,7 @@ static int handle_leader(struct handle *req, struct cursor *cursor)
 	for (i = 0; i < g->raft->configuration.n; i++) {
 		struct raft_server *server = &g->raft->configuration.servers[i];
 		if (server->id == g->raft->id && server->role == RAFT_VOTER) {
+                        tracef("handle leader - dispatch to %llu", server->id);
 			raft_leader(req->gateway->raft, &id, &address);
 			break;
 		}
@@ -153,6 +209,7 @@ static int handle_leader(struct handle *req, struct cursor *cursor)
 
 static int handle_client(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle client");
 	START(client, welcome);
 	response.heartbeat_timeout = req->gateway->config->heartbeat_timeout;
 	SUCCESS(welcome, WELCOME);
@@ -161,25 +218,32 @@ static int handle_client(struct handle *req, struct cursor *cursor)
 
 static int handle_open(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle open");
 	struct gateway *g = req->gateway;
 	struct db *db;
 	int rc;
 	START(open, db);
 	if (g->leader != NULL) {
+                tracef("already open");
 		failure(req, SQLITE_BUSY,
 			"a database for this connection is already open");
 		return 0;
 	}
 	rc = registry__db_get(g->registry, request.filename, &db);
 	if (rc != 0) {
+                tracef("registry db get failed %d", rc);
 		return rc;
 	}
 	g->leader = sqlite3_malloc(sizeof *g->leader);
 	if (g->leader == NULL) {
+                tracef("malloc failed");
 		return DQLITE_NOMEM;
 	}
 	rc = leader__init(g->leader, db, g->raft);
 	if (rc != 0) {
+                tracef("leader init failed %d", rc);
+		sqlite3_free(g->leader);
+		g->leader = NULL;
 		return rc;
 	}
 	response.id = 0;
@@ -189,6 +253,7 @@ static int handle_open(struct handle *req, struct cursor *cursor)
 
 static int handle_prepare(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle prepare");
 	struct gateway *g = req->gateway;
 	struct stmt *stmt;
 	const char *tail;
@@ -197,12 +262,14 @@ static int handle_prepare(struct handle *req, struct cursor *cursor)
 	LOOKUP_DB(request.db_id);
 	rc = stmt__registry_add(&g->stmts, &stmt);
 	if (rc != 0) {
+                tracef("handle prepare registry add failed %d", rc);
 		return rc;
 	}
 	assert(stmt != NULL);
 	rc = sqlite3_prepare_v2(g->leader->conn, request.sql, -1, &stmt->stmt,
 				&tail);
 	if (rc != SQLITE_OK) {
+                tracef("handle prepare sqlite prepare failed %d", rc);
 		failure(req, rc, sqlite3_errmsg(g->leader->conn));
 		return 0;
 	}
@@ -256,6 +323,7 @@ static void leader_exec_cb(struct exec *exec, int status)
 
 static int handle_exec(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle exec");
 	struct gateway *g = req->gateway;
 	struct stmt *stmt;
 	int rv;
@@ -263,9 +331,11 @@ static int handle_exec(struct handle *req, struct cursor *cursor)
 	CHECK_LEADER(req);
 	LOOKUP_DB(request.db_id);
 	LOOKUP_STMT(request.stmt_id);
+	FAIL_IF_CHECKPOINTING;
 	(void)response;
 	rv = bind__params(stmt->stmt, cursor);
 	if (rv != 0) {
+                tracef("handle exec bind failed %d", rv);
 		failure(req, rv, "bind parameters");
 		return 0;
 	}
@@ -273,6 +343,7 @@ static int handle_exec(struct handle *req, struct cursor *cursor)
 	g->stmt = stmt->stmt;
 	rv = leader__exec(g->leader, &g->exec, stmt->stmt, leader_exec_cb);
 	if (rv != 0) {
+                tracef("handle exec leader exec failed %d", rv);
 		return rv;
 	}
 	return 0;
@@ -318,6 +389,7 @@ done:
 
 static void query_barrier_cb(struct barrier *barrier, int status)
 {
+	tracef("query barrier cb status:%d", status);
 	struct gateway *g = barrier->data;
 	struct handle *handle = g->req;
 	sqlite3_stmt *stmt = g->stmt;
@@ -329,6 +401,10 @@ static void query_barrier_cb(struct barrier *barrier, int status)
 	g->req = NULL;
 
 	if (status != 0) {
+		if (g->stmt_finalize) {
+			sqlite3_finalize(stmt);
+			g->stmt_finalize = false;
+		}
 		failure(handle, status, "barrier error");
 		return;
 	}
@@ -338,6 +414,7 @@ static void query_barrier_cb(struct barrier *barrier, int status)
 
 static int handle_query(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle query");
 	struct gateway *g = req->gateway;
 	struct stmt *stmt;
 	int rv;
@@ -345,9 +422,11 @@ static int handle_query(struct handle *req, struct cursor *cursor)
 	CHECK_LEADER(req);
 	LOOKUP_DB(request.db_id);
 	LOOKUP_STMT(request.stmt_id);
+	FAIL_IF_CHECKPOINTING;
 	(void)response;
 	rv = bind__params(stmt->stmt, cursor);
 	if (rv != 0) {
+                tracef("handle query bind failed %d", rv);
 		failure(req, rv, sqlite3_errmsg(g->leader->conn));
 		return 0;
 	}
@@ -355,6 +434,7 @@ static int handle_query(struct handle *req, struct cursor *cursor)
 	g->stmt = stmt->stmt;
 	rv = leader__barrier(g->leader, &g->barrier, query_barrier_cb);
 	if (rv != 0) {
+                tracef("handle query leader barrier failed %d", rv);
 		g->req = NULL;
 		g->stmt = NULL;
 		return rv;
@@ -364,6 +444,7 @@ static int handle_query(struct handle *req, struct cursor *cursor)
 
 static int handle_finalize(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle finalize");
 	struct stmt *stmt;
 	int rv;
 	START(finalize, empty);
@@ -371,6 +452,7 @@ static int handle_finalize(struct handle *req, struct cursor *cursor)
 	LOOKUP_STMT(request.stmt_id);
 	rv = stmt__registry_del(&req->gateway->stmts, stmt);
 	if (rv != 0) {
+                tracef("handle finalize registry del failed %d", rv);
 		failure(req, rv, "finalize statement");
 		return 0;
 	}
@@ -382,6 +464,7 @@ static void handle_exec_sql_next(struct handle *req, struct cursor *cursor);
 
 static void handle_exec_sql_cb(struct exec *exec, int status)
 {
+        tracef("handle exec sql cb status %d", status);
 	struct gateway *g = exec->data;
 	struct handle *req = g->req;
 
@@ -399,10 +482,10 @@ static void handle_exec_sql_cb(struct exec *exec, int status)
 
 static void handle_exec_sql_next(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle exec sql next");
 	struct gateway *g = req->gateway;
 	struct response_result response;
 	const char *tail;
-	sqlite3_stmt *stmt;
 	int rv;
 
 	if (g->sql == NULL || strcmp(g->sql, "") == 0) {
@@ -414,21 +497,20 @@ static void handle_exec_sql_next(struct handle *req, struct cursor *cursor)
 		g->stmt = NULL;
 	}
 
-	rv = sqlite3_prepare_v2(g->leader->conn, g->sql, -1, &stmt, &tail);
+	/* g->stmt will be set to NULL by sqlite when an error occurs. */
+	rv = sqlite3_prepare_v2(g->leader->conn, g->sql, -1, &g->stmt, &tail);
 	if (rv != SQLITE_OK) {
 		failure(req, rv, sqlite3_errmsg(g->leader->conn));
 		goto done;
 	}
 
-	if (stmt == NULL) {
+	if (g->stmt == NULL) {
 		goto success;
 	}
 
-	g->stmt = stmt;
-
 	/* TODO: what about bindings for multi-statement SQL text? */
 	if (cursor != NULL) {
-		rv = bind__params(stmt, cursor);
+		rv = bind__params(g->stmt, cursor);
 		if (rv != SQLITE_OK) {
 			failure(req, rv, sqlite3_errmsg(g->leader->conn));
 			goto done_after_prepare;
@@ -447,6 +529,7 @@ static void handle_exec_sql_next(struct handle *req, struct cursor *cursor)
 	return;
 
 success:
+        tracef("handle exec sql next success");
 	if (g->stmt != NULL) {
 		fill_result(g, &response);
 	}
@@ -464,10 +547,12 @@ done:
 
 static int handle_exec_sql(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle exec sql");
 	struct gateway *g = req->gateway;
 	START(exec_sql, result);
 	CHECK_LEADER(req);
 	LOOKUP_DB(request.db_id);
+	FAIL_IF_CHECKPOINTING;
 	(void)response;
 	assert(g->req == NULL);
 	assert(g->sql == NULL);
@@ -479,21 +564,25 @@ static int handle_exec_sql(struct handle *req, struct cursor *cursor)
 
 static int handle_query_sql(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle query sql");
 	struct gateway *g = req->gateway;
 	const char *tail;
 	int rv;
 	START(query_sql, rows);
 	CHECK_LEADER(req);
 	LOOKUP_DB(request.db_id);
+	FAIL_IF_CHECKPOINTING;
 	(void)response;
 	rv = sqlite3_prepare_v2(g->leader->conn, request.sql, -1, &g->stmt,
 				&tail);
 	if (rv != SQLITE_OK) {
+                tracef("handle query sql prepare failed %d", rv);
 		failure(req, rv, sqlite3_errmsg(g->leader->conn));
 		return 0;
 	}
 	rv = bind__params(g->stmt, cursor);
 	if (rv != 0) {
+                tracef("handle query sql bind failed %d", rv);
 		sqlite3_finalize(g->stmt);
 		g->stmt = NULL;
 		failure(req, rv, sqlite3_errmsg(g->leader->conn));
@@ -503,8 +592,11 @@ static int handle_query_sql(struct handle *req, struct cursor *cursor)
 	g->req = req;
 	rv = leader__barrier(g->leader, &g->barrier, query_barrier_cb);
 	if (rv != 0) {
+                tracef("barrier failed %d", rv);
 		g->req = NULL;
+		sqlite3_finalize(g->stmt);
 		g->stmt = NULL;
+		g->stmt_finalize = false;
 		return rv;
 	}
 	return 0;
@@ -512,6 +604,7 @@ static int handle_query_sql(struct handle *req, struct cursor *cursor)
 
 static int handle_interrupt(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle interrupt");
 	struct gateway *g = req->gateway;
 	START(interrupt, empty);
 
@@ -528,21 +621,6 @@ static int handle_interrupt(struct handle *req, struct cursor *cursor)
 	return 0;
 }
 
-/* Translate a raft error to a dqlite one. */
-static int translateRaftErrCode(int code)
-{
-	switch (code) {
-		case RAFT_NOTLEADER:
-			return SQLITE_IOERR_NOT_LEADER;
-		case RAFT_LEADERSHIPLOST:
-			return SQLITE_IOERR_LEADERSHIP_LOST;
-		case RAFT_CANTCHANGE:
-			return SQLITE_BUSY;
-		default:
-			return SQLITE_ERROR;
-	}
-}
-
 struct change
 {
 	struct gateway *gateway;
@@ -551,6 +629,7 @@ struct change
 
 static void raftChangeCb(struct raft_change *change, int status)
 {
+        tracef("raft change cb status %d", status);
 	struct change *r = change->data;
 	struct gateway *g = r->gateway;
 	struct handle *req = g->req;
@@ -567,6 +646,7 @@ static void raftChangeCb(struct raft_change *change, int status)
 
 static int handle_add(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle add");
 	struct gateway *g = req->gateway;
 	struct change *r;
 	int rv;
@@ -581,38 +661,24 @@ static int handle_add(struct handle *req, struct cursor *cursor)
 	}
 	r->gateway = g;
 	r->req.data = r;
+	g->req = req;
 
 	rv = raft_add(g->raft, &r->req, request.id, request.address,
 		      raftChangeCb);
 	if (rv != 0) {
+                tracef("raft add failed %d", rv);
+		g->req = NULL;
 		sqlite3_free(r);
 		failure(req, translateRaftErrCode(rv), raft_strerror(rv));
 		return 0;
 	}
-	g->req = req;
 
 	return 0;
 }
 
-/* Translate a dqlite role code to its raft equivalent. */
-static int translateDqliteRole(int role)
-{
-	switch (role) {
-		case DQLITE_VOTER:
-			return RAFT_VOTER;
-		case DQLITE_STANDBY:
-			return RAFT_STANDBY;
-		case DQLITE_SPARE:
-			return RAFT_SPARE;
-		default:
-			/* For backward compat with clients that don't set a
-			 * role. */
-			return DQLITE_VOTER;
-	}
-}
-
 static int handle_assign(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle assign");
 	struct gateway *g = req->gateway;
 	struct change *r;
 	uint64_t role = DQLITE_VOTER;
@@ -627,31 +693,36 @@ static int handle_assign(struct handle *req, struct cursor *cursor)
 	if (cursor->cap > 0) {
 		rv = uint64__decode(cursor, &role);
 		if (rv != 0) {
+                        tracef("handle assign promote rv %d", rv);
 			return rv;
 		}
 	}
 
 	r = sqlite3_malloc(sizeof *r);
 	if (r == NULL) {
+                tracef("malloc failed");
 		return DQLITE_NOMEM;
 	}
 	r->gateway = g;
 	r->req.data = r;
+	g->req = req;
 
 	rv = raft_assign(g->raft, &r->req, request.id,
 			 translateDqliteRole((int)role), raftChangeCb);
 	if (rv != 0) {
+                tracef("raft_assign failed %d", rv);
+		g->req = NULL;
 		sqlite3_free(r);
 		failure(req, translateRaftErrCode(rv), raft_strerror(rv));
 		return 0;
 	}
-	g->req = req;
 
 	return 0;
 }
 
 static int handle_remove(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle remove");
 	struct gateway *g = req->gateway;
 	struct change *r;
 	int rv;
@@ -662,18 +733,21 @@ static int handle_remove(struct handle *req, struct cursor *cursor)
 
 	r = sqlite3_malloc(sizeof *r);
 	if (r == NULL) {
+                tracef("malloc failed");
 		return DQLITE_NOMEM;
 	}
 	r->gateway = g;
 	r->req.data = r;
+	g->req = req;
 
 	rv = raft_remove(g->raft, &r->req, request.id, raftChangeCb);
 	if (rv != 0) {
+                tracef("raft_remote failed %d", rv);
+		g->req = NULL;
 		sqlite3_free(r);
 		failure(req, translateRaftErrCode(rv), raft_strerror(rv));
 		return 0;
 	}
-	g->req = req;
 
 	return 0;
 }
@@ -718,10 +792,12 @@ oom:
 
 static int handle_dump(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle dump");
+	bool err = true;
 	struct gateway *g = req->gateway;
 	sqlite3_vfs *vfs;
 	void *cur;
-	char filename[1024];
+	char filename[1024] = {0};
 	void *data;
 	size_t n;
 	uint8_t *page;
@@ -741,6 +817,7 @@ static int handle_dump(struct handle *req, struct cursor *cursor)
 	vfs = sqlite3_vfs_find(g->config->name);
 	rv = VfsSnapshot(vfs, request.filename, &data, &n);
 	if (rv != 0) {
+                tracef("dump failed");
 		failure(req, rv, "failed to dump database");
 		return 0;
 	}
@@ -770,43 +847,38 @@ static int handle_dump(struct handle *req, struct cursor *cursor)
 
 	rv = dumpFile(request.filename, database, n_database, req->buffer);
 	if (rv != 0) {
+                tracef("dump failed");
 		failure(req, rv, "failed to dump database");
-		return 0;
+		goto out_free_data;
 	}
 
-	strcpy(filename, request.filename);
-	strcat(filename, "-wal");
+	/* filename is zero inited and initially we allow only writing 1024 - 4
+	 * - 1 bytes to it, so after strncpy filename will be zero-terminated
+	 * and will not have overflowed. strcat adds the 4 byte suffix and
+	 * also zero terminates the resulting string. */
+	const char * wal_suffix = "-wal";
+	strncpy(filename, request.filename, sizeof(filename) - strlen(wal_suffix) - 1);
+	strcat(filename, wal_suffix);
 	rv = dumpFile(filename, wal, n_wal, req->buffer);
 	if (rv != 0) {
+                tracef("wal dump failed");
 		failure(req, rv, "failed to dump wal file");
-		return 0;
+		goto out_free_data;
 	}
 
+	err = false;
+
+out_free_data:
 	if (data != NULL) {
 		raft_free(data);
 	}
 
-	req->cb(req, 0, DQLITE_RESPONSE_FILES);
+	if (!err)
+		req->cb(req, 0, DQLITE_RESPONSE_FILES);
 
 	return 0;
 }
 
-/* Translate a raft role code to its dqlite equivalent. */
-static int translateRaftRole(int role)
-{
-	switch (role) {
-		case RAFT_VOTER:
-			return DQLITE_VOTER;
-		case RAFT_STANDBY:
-			return DQLITE_STANDBY;
-		case RAFT_SPARE:
-			return DQLITE_SPARE;
-		default:
-			assert(0);
-			return -1;
-	}
-}
-
 static int encodeServer(struct gateway *g,
 			unsigned i,
 			struct buffer *buffer,
@@ -849,6 +921,7 @@ static int encodeServer(struct gateway *g,
 
 static int handle_cluster(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle cluster");
 	struct gateway *g = req->gateway;
 	unsigned i;
 	void *cur;
@@ -863,6 +936,7 @@ static int handle_cluster(struct handle *req, struct cursor *cursor)
 	for (i = 0; i < response.n; i++) {
 		rv = encodeServer(g, i, req->buffer, (int)request.format);
 		if (rv != 0) {
+                        tracef("encode failed");
 			failure(req, rv, "failed to encode server");
 			return 0;
 		}
@@ -881,6 +955,7 @@ void raftTransferCb(struct raft_transfer *r)
 	g->req = NULL;
 	sqlite3_free(r);
 	if (g->raft->state == RAFT_LEADER) {
+                tracef("transfer failed");
 		failure(req, DQLITE_ERROR, "leadership transfer failed");
 	} else {
 		SUCCESS(empty, EMPTY);
@@ -889,6 +964,7 @@ void raftTransferCb(struct raft_transfer *r)
 
 static int handle_transfer(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle transfer");
 	struct gateway *g = req->gateway;
 	struct raft_transfer *r;
 	int rv;
@@ -899,26 +975,31 @@ static int handle_transfer(struct handle *req, struct cursor *cursor)
 
 	r = sqlite3_malloc(sizeof *r);
 	if (r == NULL) {
+                tracef("malloc failed");
 		return DQLITE_NOMEM;
 	}
 	r->data = g;
+	g->req = req;
 
 	rv = raft_transfer(g->raft, r, request.id, raftTransferCb);
 	if (rv != 0) {
+                tracef("raft_transfer failed %d", rv);
+		g->req = NULL;
 		sqlite3_free(r);
 		failure(req, translateRaftErrCode(rv), raft_strerror(rv));
 		return 0;
 	}
-	g->req = req;
 
 	return 0;
 }
 
 static int handle_describe(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle describe");
 	struct gateway *g = req->gateway;
 	START(describe, metadata);
 	if (request.format != DQLITE_REQUEST_DESCRIBE_FORMAT_V0) {
+                tracef("bad format");
 		failure(req, SQLITE_PROTOCOL, "bad format version");
 	}
 	response.failure_domain = g->config->failure_domain;
@@ -929,6 +1010,7 @@ static int handle_describe(struct handle *req, struct cursor *cursor)
 
 static int handle_weight(struct handle *req, struct cursor *cursor)
 {
+        tracef("handle weight");
 	struct gateway *g = req->gateway;
 	START(weight, empty);
 	g->config->weight = request.weight;
@@ -943,6 +1025,7 @@ int gateway__handle(struct gateway *g,
 		    struct buffer *buffer,
 		    handle_cb cb)
 {
+        tracef("gateway handle");
 	int rc = 0;
 
 	/* Check if there is a request in progress. */
@@ -955,6 +1038,7 @@ int gateway__handle(struct gateway *g,
 		}
 		if (g->req->type == DQLITE_REQUEST_EXEC ||
 		    g->req->type == DQLITE_REQUEST_EXEC_SQL) {
+                        tracef("gateway handle - BUSY");
 			return SQLITE_BUSY;
 		}
 		assert(0);
@@ -981,10 +1065,12 @@ int gateway__resume(struct gateway *g, bool *finished)
 {
 	if (g->req == NULL || (g->req->type != DQLITE_REQUEST_QUERY &&
 			       g->req->type != DQLITE_REQUEST_QUERY_SQL)) {
+                tracef("gateway resume - finished");
 		*finished = true;
 		return 0;
 	}
 	assert(g->stmt != NULL);
+        tracef("gateway resume - not finished");
 	*finished = false;
 	query_batch(g->stmt, g->req);
 	return 0;
