@@ -10,6 +10,7 @@
 #include "lib/assert.h"
 #include "logger.h"
 #include "protocol.h"
+#include "tracing.h"
 #include "translate.h"
 #include "transport.h"
 #include "utils.h"
@@ -180,7 +181,8 @@ static int ipParse(const char *address, struct sockaddr_in *addr)
 	int rv;
 
 	/* TODO: turn this poor man parsing into proper one */
-	strcpy(buf, address);
+	strncpy(buf, address, sizeof(buf)-1);
+	buf[sizeof(buf)-1] = '\0';
 	host = strtok(buf, colon);
 	port = strtok(NULL, ":");
 	if (port == NULL) {
@@ -273,7 +275,8 @@ int dqlite_node_set_bind_address(dqlite_node *t, const char *address)
 	}
 
 	if (domain == AF_INET) {
-		t->bind_address = sqlite3_malloc((int)strlen(address));
+		int sz = ((int)strlen(address)) + 1; /* Room for '\0' */
+		t->bind_address = sqlite3_malloc(sz);
 		if (t->bind_address == NULL) {
 			close(fd);
 			return DQLITE_NOMEM;
@@ -697,17 +700,22 @@ int dqlite_node_start(dqlite_node *t)
 {
 	int rv;
 
+	dqliteTracingMaybeEnable(true);
+	tracef("dqlite node start");
 	rv = maybeBootstrap(t, t->config.id, t->config.address);
 	if (rv != 0) {
+		tracef("bootstrap failed %d", rv);
 		goto err;
 	}
 
 	rv = pthread_create(&t->thread, 0, &taskStart, t);
 	if (rv != 0) {
+		tracef("pthread create failed %d", rv);
 		goto err;
 	}
 
 	if (!taskReady(t)) {
+		tracef("!taskReady");
 		rv = DQLITE_ERROR;
 		goto err;
 	}
@@ -720,6 +728,7 @@ err:
 
 int dqlite_node_stop(dqlite_node *d)
 {
+	tracef("dqlite node stop");
 	void *result;
 	int rv;
 
@@ -748,6 +757,7 @@ int dqlite_node_recover(dqlite_node *n,
 			struct dqlite_node_info infos[],
 			int n_info)
 {
+    tracef("dqlite node recover");
     int i;
     int ret;
 
@@ -801,6 +811,7 @@ int dqlite_node_recover_ext(dqlite_node *n,
 			struct dqlite_node_info_ext infos[],
 			int n_info)
 {
+    	tracef("dqlite node recover ext");
 	struct raft_configuration configuration;
 	int i;
 	int rv;
@@ -836,6 +847,7 @@ out:
 
 dqlite_node_id dqlite_generate_node_id(const char *address)
 {
+	tracef("generate node id");
 	struct timespec ts;
 	int rv;
 	unsigned long long n;
