@@ -13,6 +13,7 @@
 #include "lib/assert.h"
 
 #include "format.h"
+#include "tracing.h"
 #include "vfs.h"
 
 /* Byte order */
@@ -75,7 +76,9 @@ struct vfsFrame
 	uint8_t *page; /* Content of the page. */
 };
 
-/* WAL-specific content */
+/* WAL-specific content.
+ * Watch out when changing the members of this struct, see
+ * comment in `formatWalChecksumBytes`. */
 struct vfsWal
 {
 	uint8_t hdr[VFS__WAL_HEADER_SIZE]; /* Header. */
@@ -142,13 +145,31 @@ static uint32_t vfsFlip32(uint32_t v)
 /* Load a 16-bit number stored in big-endian representation. */
 static uint32_t vfsGet16(const uint8_t *buf)
 {
-	return vfsFlip16(*(const uint16_t *)buf);
+	union {
+		uint16_t u;
+		uint8_t v[2];
+	} s;
+
+	s.v[0] = buf[0];
+	s.v[1] = buf[1];
+
+	return vfsFlip16(s.u);
 }
 
 /* Load a 32-bit number stored in big-endian representation. */
 static uint32_t vfsGet32(const uint8_t *buf)
 {
-	return vfsFlip32(*(const uint32_t *)buf);
+	union {
+		uint32_t u;
+		uint8_t v[4];
+	} s;
+
+	s.v[0] = buf[0];
+	s.v[1] = buf[1];
+	s.v[2] = buf[2];
+	s.v[3] = buf[3];
+
+	return vfsFlip32(s.u);
 }
 
 /* Store a 32-bit number in big endian format */
@@ -173,9 +194,11 @@ static void vfsChecksum(
     uint32_t out[2]       /* OUT: Final checksum value output */
 )
 {
+	assert((((uintptr_t)data) % sizeof(uint32_t)) == 0);
+
 	uint32_t s1, s2;
-	uint32_t *cur = (uint32_t *)data;
-	uint32_t *end = (uint32_t *)&data[n];
+	uint32_t *cur = (uint32_t *)__builtin_assume_aligned(data, sizeof(uint32_t));
+	uint32_t *end = (uint32_t *)__builtin_assume_aligned(&data[n], sizeof(uint32_t));
 
 	if (in) {
 		s1 = in[0];
@@ -1381,19 +1404,22 @@ static void vfsAmendWalIndexHeader(struct vfsDatabase *d)
 	assert(shm->n_regions > 0);
 	index = shm->regions[0];
 
-	assert(*(uint32_t *)(&index[0]) == VFS__WAL_VERSION); /* iVersion */
+	/* index is an alias for shm->regions[0] which is a void* that points to
+	 * memory allocated by `sqlite3_malloc64` and has the required alignment */
+	assert(*(uint32_t *)(__builtin_assume_aligned(&index[0], sizeof(uint32_t)))
+			     == VFS__WAL_VERSION); /* iVersion */
 	assert(index[12] == 1);                               /* isInit */
 	assert(index[13] == VFS__BIGENDIAN);                  /* bigEndCksum */
 
-	*(uint32_t *)(&index[16]) = wal->n_frames;
-	*(uint32_t *)(&index[20]) = n_pages;
-	*(uint32_t *)(&index[24]) = frame_checksum[0];
-	*(uint32_t *)(&index[28]) = frame_checksum[1];
+	*(uint32_t *)(__builtin_assume_aligned(&index[16], sizeof(uint32_t))) = wal->n_frames;
+	*(uint32_t *)(__builtin_assume_aligned(&index[20], sizeof(uint32_t))) = n_pages;
+	*(uint32_t *)(__builtin_assume_aligned(&index[24], sizeof(uint32_t))) = frame_checksum[0];
+	*(uint32_t *)(__builtin_assume_aligned(&index[28], sizeof(uint32_t))) = frame_checksum[1];
 
 	vfsChecksum(index, 40, checksum, checksum);
 
-	*(uint32_t *)(&index[40]) = checksum[0];
-	*(uint32_t *)(&index[44]) = checksum[1];
+	*(uint32_t *)__builtin_assume_aligned(&index[40], sizeof(uint32_t)) = checksum[0];
+	*(uint32_t *)__builtin_assume_aligned(&index[44], sizeof(uint32_t)) = checksum[1];
 
 	/* Update the second copy of the first part of the WAL index header. */
 	memcpy(index + VFS__WAL_INDEX_HEADER_SIZE, index,
@@ -1542,6 +1568,8 @@ static int vfsShmLock(struct vfsShm *s, int ofst, int n, int flags)
 		/* No shared or exclusive lock must be held in the region. */
 		for (i = ofst; i < ofst + n; i++) {
 			if (s->shared[i] > 0 || s->exclusive[i] > 0) {
+			        tracef("EXCLUSIVE lock contention ofst:%d n:%d exclusive[%d]=%d shared[%d]=%d",
+			                ofst, n, i, s->exclusive[i], i, s->shared[i]);
 				return SQLITE_BUSY;
 			}
 		}
@@ -1554,6 +1582,8 @@ static int vfsShmLock(struct vfsShm *s, int ofst, int n, int flags)
 		/* No exclusive lock must be held in the region. */
 		for (i = ofst; i < ofst + n; i++) {
 			if (s->exclusive[i] > 0) {
+			        tracef("SHARED lock contention ofst:%d n:%d exclusive[%d]=%d shared[%d]=%d",
+			                ofst, n, i, s->exclusive[i], i, s->shared[i]);
 				return SQLITE_BUSY;
 			}
 		}
@@ -1606,10 +1636,12 @@ static void vfsWalRollbackIfUncommitted(struct vfsWal *w)
 		return;
 	}
 
+        tracef("rollback n_tx:%d", w->n_tx);
 	last = w->tx[w->n_tx - 1];
 	commit = vfsFrameGetDatabaseSize(last);
 
 	if (commit > 0) {
+                tracef("rollback commit:%u", commit);
 		return;
 	}
 
@@ -2080,11 +2112,13 @@ static int vfsInit(struct sqlite3_vfs *vfs, const char *name)
 
 int VfsInit(struct sqlite3_vfs *vfs, const char *name)
 {
+	tracef("vfs init");
 	return vfsInit(vfs, name);
 }
 
 void VfsClose(struct sqlite3_vfs *vfs)
 {
+	tracef("vfs close");
 	struct vfs *v = vfs->pAppData;
 	vfsDestroy(v);
 	sqlite3_free(v);
@@ -2139,6 +2173,7 @@ int VfsPoll(sqlite3_vfs *vfs,
 	    dqlite_vfs_frame **frames,
 	    unsigned *n)
 {
+	tracef("vfs poll filename:%s", filename);
 	struct vfs *v;
 	struct vfsDatabase *database;
 	struct vfsShm *shm;
@@ -2149,6 +2184,7 @@ int VfsPoll(sqlite3_vfs *vfs,
 	database = vfsDatabaseLookup(v, filename);
 
 	if (database == NULL) {
+		tracef("not found");
 		return DQLITE_ERROR;
 	}
 
@@ -2163,6 +2199,7 @@ int VfsPoll(sqlite3_vfs *vfs,
 
 	rv = vfsWalPoll(wal, frames, n);
 	if (rv != 0) {
+		tracef("wal poll failed %d", rv);
 		return rv;
 	}
 
@@ -2170,6 +2207,7 @@ int VfsPoll(sqlite3_vfs *vfs,
 	if (*n > 0) {
 		rv = vfsShmLock(shm, 0, 1, SQLITE_SHM_EXCLUSIVE);
 		if (rv != 0) {
+			tracef("shm lock failed %d", rv);
 			return rv;
 		}
 		vfsAmendWalIndexHeader(database);
@@ -2181,13 +2219,16 @@ int VfsPoll(sqlite3_vfs *vfs,
 /* Return the salt-1 field stored in the WAL header.*/
 static uint32_t vfsWalGetSalt1(struct vfsWal *w)
 {
-	return *(uint32_t *)(&w->hdr[16]);
+
+	/* `hdr` field is pointer aligned, cast is safe */
+	return *(uint32_t *)__builtin_assume_aligned(&w->hdr[16], sizeof(uint32_t));
 }
 
 /* Return the salt-2 field stored in the WAL header.*/
 static uint32_t vfsWalGetSalt2(struct vfsWal *w)
 {
-	return *(uint32_t *)(&w->hdr[20]);
+	/* `hdr` field is pointer aligned, cast is safe */
+	return *(uint32_t *)__builtin_assume_aligned(&w->hdr[20], sizeof(uint32_t));
 }
 
 /* Return the checksum-1 field stored in the WAL header.*/
@@ -2346,6 +2387,7 @@ int VfsApply(sqlite3_vfs *vfs,
 	     unsigned long *page_numbers,
 	     void *frames)
 {
+	tracef("vfs apply filename %s n %u", filename, n);
 	struct vfs *v;
 	struct vfsDatabase *database;
 	struct vfsWal *wal;
@@ -2369,6 +2411,7 @@ int VfsApply(sqlite3_vfs *vfs,
 
 	rv = vfsWalAppend(wal, database->n_pages, n, page_numbers, frames);
 	if (rv != 0) {
+		tracef("wal append failed rv:%d n_pages:%u n:%u", rv, database->n_pages, n);
 		return rv;
 	}
 
@@ -2395,6 +2438,7 @@ int VfsApply(sqlite3_vfs *vfs,
 
 int VfsAbort(sqlite3_vfs *vfs, const char *filename)
 {
+	tracef("vfs abort filename %s", filename);
 	struct vfs *v;
 	struct vfsDatabase *database;
 	int rv;
@@ -2404,6 +2448,7 @@ int VfsAbort(sqlite3_vfs *vfs, const char *filename)
 
 	rv = vfsShmUnlock(&database->shm, 0, 1, SQLITE_SHM_EXCLUSIVE);
 	if (rv != 0) {
+		tracef("shm unlock failed %d", rv);
 		return rv;
 	}
 
@@ -2465,6 +2510,7 @@ static void vfsWalSnapshot(struct vfsWal *w, uint8_t **cursor)
 
 int VfsSnapshot(sqlite3_vfs *vfs, const char *filename, void **data, size_t *n)
 {
+	tracef("vfs snapshot filename %s", filename);
 	struct vfs *v;
 	struct vfsDatabase *database;
 	struct vfsWal *wal;
@@ -2474,12 +2520,14 @@ int VfsSnapshot(sqlite3_vfs *vfs, const char *filename, void **data, size_t *n)
 	database = vfsDatabaseLookup(v, filename);
 
 	if (database == NULL) {
+		tracef("not found");
 		*data = NULL;
 		*n = 0;
 		return 0;
 	}
 
 	if (database->n_pages != vfsDatabaseGetNumberOfPages(database)) {
+		tracef("corrupt");
 		return SQLITE_CORRUPT;
 	}
 
@@ -2489,6 +2537,7 @@ int VfsSnapshot(sqlite3_vfs *vfs, const char *filename, void **data, size_t *n)
 	/* TODO: we should fix the tests and use sqlite3_malloc instead. */
 	*data = raft_malloc(*n);
 	if (*data == NULL) {
+		tracef("malloc");
 		return DQLITE_NOMEM;
 	}
 
@@ -2590,8 +2639,8 @@ static int vfsWalRestore(struct vfsWal *w,
 			unsigned j;
 			for (j = 0; j < i; j++) {
 				vfsFrameDestroy(frames[j]);
-				goto oom_after_frames_alloc;
 			}
+			goto oom_after_frames_alloc;
 		}
 		frames[i] = frame;
 
@@ -2621,6 +2670,7 @@ int VfsRestore(sqlite3_vfs *vfs,
 	       const void *data,
 	       size_t n)
 {
+	tracef("vfs restore filename %s size %zd", filename, n);
 	struct vfs *v;
 	struct vfsDatabase *database;
 	struct vfsWal *wal;
@@ -2637,12 +2687,14 @@ int VfsRestore(sqlite3_vfs *vfs,
 	/* Truncate any existing content. */
 	rv = vfsWalTruncate(wal, 0);
 	if (rv != 0) {
+        	tracef("wal truncate failed %d", rv);
 		return rv;
 	}
 
 	/* Restore the content of the main database and of the WAL. */
 	rv = vfsDatabaseRestore(database, data, n);
 	if (rv != 0) {
+        	tracef("database restore failed %d", rv);
 		return rv;
 	}
 
@@ -2651,6 +2703,7 @@ int VfsRestore(sqlite3_vfs *vfs,
 
 	rv = vfsWalRestore(wal, data + offset, n - offset, page_size);
 	if (rv != 0) {
+        	tracef("wal restore failed %d", rv);
 		return rv;
 	}
 
