@@ -1698,7 +1698,7 @@ munit_test_runner_run(MunitTestRunner* runner) {
 }
 
 static void
-munit_print_help(int argc, char* const argv[MUNIT_ARRAY_PARAM(argc + 1)], void* user_data, const MunitArgument arguments[]) {
+munit_print_help(int argc, char* const argv[MUNIT_ARRAY_PARAM(argc)], void* user_data, const MunitArgument arguments[]) {
   const MunitArgument* arg;
   (void) argc;
 
@@ -1834,7 +1834,7 @@ munit_stream_supports_ansi(FILE *stream) {
 
 int
 munit_suite_main_custom(const MunitSuite* suite, void* user_data,
-                        int argc, char* const argv[MUNIT_ARRAY_PARAM(argc + 1)],
+                        int argc, char* const argv[MUNIT_ARRAY_PARAM(argc)],
                         const MunitArgument arguments[]) {
   int result = EXIT_FAILURE;
   MunitTestRunner runner;
@@ -2051,6 +2051,6 @@ munit_suite_main_custom(const MunitSuite* suite, void* user_data,
 
 int
 munit_suite_main(const MunitSuite* suite, void* user_data,
-                 int argc, char* const argv[MUNIT_ARRAY_PARAM(argc + 1)]) {
+                 int argc, char* const argv[MUNIT_ARRAY_PARAM(argc)]) {
   return munit_suite_main_custom(suite, user_data, argc, argv, NULL);
 }
