@@ -218,6 +218,16 @@ static void handleCb(struct handle *req, int status, int type)
 		HANDLE(EXEC_SQL);                 \
 	}
 
+/* Submit a request to execute the given statement. */
+#define QUERY_SQL_SUBMIT(SQL)                       \
+	{                                           \
+		struct request_query_sql query_sql; \
+		query_sql.db_id = 0;                \
+		query_sql.sql = SQL;                \
+		ENCODE(&query_sql, query_sql);      \
+		HANDLE(QUERY_SQL);                  \
+	}
+
 /* Wait for the last request to complete */
 #define WAIT                                            \
 	{                                               \
@@ -893,7 +903,7 @@ TEST_CASE(exec, restore, NULL)
 	munit_assert_int(value.type, ==, SQLITE_INTEGER);
 	munit_assert_int(value.integer, ==, 2);
 	DECODE(&response, rows);
-	munit_assert_ulong(response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
+	munit_assert_ullong(response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
 	return MUNIT_OK;
 }
 
@@ -947,7 +957,7 @@ TEST_CASE(query, simple, NULL)
 	text__decode(f->cursor, &column);
 	munit_assert_string_equal(column, "n");
 	DECODE(&f->response, rows);
-	munit_assert_ulong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
 
 	return MUNIT_OK;
 }
@@ -978,11 +988,21 @@ TEST_CASE(query, one_row, NULL)
 	munit_assert_int(value.type, ==, SQLITE_INTEGER);
 	munit_assert_int(value.integer, ==, 666);
 	DECODE(&f->response, rows);
-	munit_assert_ulong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
 
 	return MUNIT_OK;
 }
 
+/* Calculate max amount of rows that can fit in 1 response buffer.
+ * A response buffer has _SC_PAGESIZE size.
+ * A response consists of n tuples each row_sz in size
+ * and an 8B EOF marker. */
+static unsigned max_rows_buffer(unsigned tuple_row_sz) {
+	unsigned buf_sz = sysconf(_SC_PAGESIZE);
+	unsigned eof_sz = 8;
+	return (buf_sz - eof_sz) / tuple_row_sz;
+}
+
 /* Successfully query that yields a large number of rows that need to be split
  * into several reponses. */
 TEST_CASE(query, large, NULL)
@@ -996,7 +1016,12 @@ TEST_CASE(query, large, NULL)
 	bool finished;
 	(void)params;
 	EXEC("BEGIN");
-	for (i = 0; i < 500; i++) {
+
+	/* 16 = 8B header + 8B value (int) */
+	unsigned n_rows_buffer = max_rows_buffer(16);
+	/* Insert 1 less than 2 response buffers worth of rows, otherwise we need
+	 * 3 responses, of which the last one contains no rows. */
+	for (i = 0; i < ((2 * n_rows_buffer) - 1); i++) {
 		EXEC("INSERT INTO test(n) VALUES(123)");
 	}
 	EXEC("COMMIT");
@@ -1013,14 +1038,15 @@ TEST_CASE(query, large, NULL)
 	text__decode(f->cursor, &column);
 	munit_assert_string_equal(column, "n");
 
-	for (i = 0; i < 255; i++) {
+	/* First response contains max amount of rows */
+	for (i = 0; i < n_rows_buffer; i++) {
 		DECODE_ROW(1, &value);
 		munit_assert_int(value.type, ==, SQLITE_INTEGER);
 		munit_assert_int(value.integer, ==, 123);
 	}
 
 	DECODE(&f->response, rows);
-	munit_assert_ulong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
 
 	gateway__resume(f->gateway, &finished);
 	munit_assert_false(finished);
@@ -1032,15 +1058,18 @@ TEST_CASE(query, large, NULL)
 	text__decode(f->cursor, &column);
 	munit_assert_string_equal(column, "n");
 
-	for (i = 0; i < 245; i++) {
+	/* Second, and last, response contains 1 less than maximum amount */
+	for (i = 0; i < n_rows_buffer - 1; i++) {
 		DECODE_ROW(1, &value);
 		munit_assert_int(value.type, ==, SQLITE_INTEGER);
 		munit_assert_int(value.integer, ==, 123);
 	}
 
 	DECODE(&f->response, rows);
-	munit_assert_ulong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
 
+	gateway__resume(f->gateway, &finished);
+	munit_assert_true(finished);
 	return MUNIT_OK;
 }
 
@@ -1086,7 +1115,11 @@ TEST_CASE(query, interrupt, NULL)
 	struct value value;
 	(void)params;
 	EXEC("BEGIN");
-	for (i = 0; i < 500; i++) {
+
+	/* 16 = 8B header + 8B value (int) */
+	unsigned n_rows_buffer = max_rows_buffer(16);
+	/* Insert 2 response buffers worth of rows */
+	for (i = 0; i < 2 * n_rows_buffer; i++) {
 		EXEC("INSERT INTO test(n) VALUES(123)");
 	}
 	EXEC("COMMIT");
@@ -1103,14 +1136,14 @@ TEST_CASE(query, interrupt, NULL)
 	text__decode(f->cursor, &column);
 	munit_assert_string_equal(column, "n");
 
-	for (i = 0; i < 255; i++) {
+	for (i = 0; i < n_rows_buffer; i++) {
 		DECODE_ROW(1, &value);
 		munit_assert_int(value.type, ==, SQLITE_INTEGER);
 		munit_assert_int(value.integer, ==, 123);
 	}
 
 	DECODE(&f->response, rows);
-	munit_assert_ulong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
 
 	ENCODE(&interrupt, interrupt);
 	HANDLE(INTERRUPT);
@@ -1120,6 +1153,50 @@ TEST_CASE(query, interrupt, NULL)
 	return MUNIT_OK;
 }
 
+/* Close the gateway during a large query. */
+TEST_CASE(query, largeClose, NULL)
+{
+	struct query_fixture *f = data;
+	unsigned i;
+	uint64_t stmt_id;
+	uint64_t n;
+	const char *column;
+	struct value value;
+	(void)params;
+	EXEC("BEGIN");
+
+	/* 16 = 8B header + 8B value (int) */
+	unsigned n_rows_buffer = max_rows_buffer(16);
+	/* Insert 2 response buffers worth of rows */
+	for (i = 0; i < 2 * n_rows_buffer; i++) {
+		EXEC("INSERT INTO test(n) VALUES(123)");
+	}
+	EXEC("COMMIT");
+
+	PREPARE("SELECT n FROM test");
+	f->request.db_id = 0;
+	f->request.stmt_id = stmt_id;
+	ENCODE(&f->request, query);
+	HANDLE(QUERY);
+	ASSERT_CALLBACK(0, ROWS);
+
+	uint64__decode(f->cursor, &n);
+	munit_assert_int(n, ==, 1);
+	text__decode(f->cursor, &column);
+	munit_assert_string_equal(column, "n");
+
+	for (i = 0; i < n_rows_buffer; i++) {
+		DECODE_ROW(1, &value);
+		munit_assert_int(value.type, ==, SQLITE_INTEGER);
+		munit_assert_int(value.integer, ==, 123);
+	}
+
+	DECODE(&f->response, rows);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
+
+	return MUNIT_OK;
+}
+
 /* Submit a query request right after the server has been re-elected and needs
  * to catch up with logs. */
 TEST_CASE(query, barrier, NULL)
@@ -1146,6 +1223,71 @@ TEST_CASE(query, barrier, NULL)
 	return MUNIT_OK;
 }
 
+/* Submit a query request right after the server has been re-elected and needs
+ * to catch up with logs, but close early */
+TEST_CASE(query, barrierInFlightQuery, NULL)
+{
+	struct query_fixture *f = data;
+	uint64_t stmt_id;
+	(void)params;
+
+	PREPARE("INSERT INTO test(n) VALUES(1)");
+	EXEC_SUBMIT(stmt_id);
+	CLUSTER_DEPOSE;
+	ASSERT_CALLBACK(0, FAILURE);
+
+	/* Re-elect ourselves and issue a query request */
+	CLUSTER_ELECT(0);
+
+	PREPARE("SELECT n FROM test");
+	f->request.db_id = 0;
+	f->request.stmt_id = stmt_id;
+	ENCODE(&f->request, query);
+	HANDLE(QUERY);
+	return MUNIT_OK;
+}
+
+/* Submit a query sql request right after the server has been re-elected and needs
+ * to catch up with logs, but close early */
+TEST_CASE(query, barrierInFlightQuerySql, NULL)
+{
+	struct query_fixture *f = data;
+	uint64_t stmt_id;
+	(void)params;
+
+	PREPARE("INSERT INTO test(n) VALUES(1)");
+	EXEC_SUBMIT(stmt_id);
+	CLUSTER_DEPOSE;
+	ASSERT_CALLBACK(0, FAILURE);
+
+	/* Re-elect ourselves and issue a query request */
+	CLUSTER_ELECT(0);
+
+	QUERY_SQL_SUBMIT("SELECT n FROM test");
+	return MUNIT_OK;
+}
+
+/* Submit an exec request right after the server has been re-elected and needs
+ * to catch up with logs, but close early */
+TEST_CASE(query, barrierInFlightExec, NULL)
+{
+	struct query_fixture *f = data;
+	uint64_t stmt_id;
+	(void)params;
+
+	PREPARE("INSERT INTO test(n) VALUES(1)");
+	EXEC_SUBMIT(stmt_id);
+	CLUSTER_DEPOSE;
+	ASSERT_CALLBACK(0, FAILURE);
+
+	/* Re-elect ourselves and issue an exec request */
+	CLUSTER_ELECT(0);
+
+	PREPARE("INSERT INTO test(n) VALUES(2)");
+	EXEC_SUBMIT(stmt_id);
+	return MUNIT_OK;
+}
+
 /******************************************************************************
  *
  * finalize
@@ -1291,6 +1433,118 @@ TEST_CASE(query_sql, small, NULL)
 	return MUNIT_OK;
 }
 
+/* Exec a SQL query whose result set needs multiple pages. */
+TEST_CASE(query_sql, large, NULL)
+{
+	struct query_sql_fixture *f = data;
+	(void)params;
+	unsigned i;
+	uint64_t n;
+	const char *column;
+	struct value value;
+	bool finished;
+	EXEC("BEGIN");
+
+	/* 16 = 8B header + 8B value (int) */
+	unsigned n_rows_buffer = max_rows_buffer(16);
+	/* Insert 1 less than 2 response buffers worth of rows, otherwise we need
+	 * 3 responses, of which the last one contains no rows. */
+	for (i = 0; i < ((2 * n_rows_buffer) - 1); i++) {
+		EXEC("INSERT INTO test(n) VALUES(123)");
+	}
+	EXEC("COMMIT");
+
+	f->request.db_id = 0;
+	f->request.sql = "SELECT n FROM test";
+	ENCODE(&f->request, query_sql);
+	HANDLE(QUERY_SQL);
+	ASSERT_CALLBACK(0, ROWS);
+
+	uint64__decode(f->cursor, &n);
+	munit_assert_int(n, ==, 1);
+	text__decode(f->cursor, &column);
+	munit_assert_string_equal(column, "n");
+
+	/* First response contains max amount of rows */
+	for (i = 0; i < n_rows_buffer; i++) {
+		DECODE_ROW(1, &value);
+		munit_assert_int(value.type, ==, SQLITE_INTEGER);
+		munit_assert_int(value.integer, ==, 123);
+	}
+
+	DECODE(&f->response, rows);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
+
+	gateway__resume(f->gateway, &finished);
+	munit_assert_false(finished);
+
+	ASSERT_CALLBACK(0, ROWS);
+
+	uint64__decode(f->cursor, &n);
+	munit_assert_int(n, ==, 1);
+	text__decode(f->cursor, &column);
+	munit_assert_string_equal(column, "n");
+
+	/* Second, and last, response contains 1 less than maximum amount */
+	for (i = 0; i < n_rows_buffer - 1; i++) {
+		DECODE_ROW(1, &value);
+		munit_assert_int(value.type, ==, SQLITE_INTEGER);
+		munit_assert_int(value.integer, ==, 123);
+	}
+
+	DECODE(&f->response, rows);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_DONE);
+
+	gateway__resume(f->gateway, &finished);
+	munit_assert_true(finished);
+	return MUNIT_OK;
+}
+
+/* Exec a SQL query whose result set needs multiple pages and close before
+ * receiving the full result set. */
+TEST_CASE(query_sql, largeClose, NULL)
+{
+	struct query_sql_fixture *f = data;
+	(void)params;
+	unsigned i;
+	uint64_t n;
+	const char *column;
+	struct value value;
+	EXEC("BEGIN");
+
+	/* 16 = 8B header + 8B value (int) */
+	unsigned n_rows_buffer = max_rows_buffer(16);
+	/* Insert 1 less than 2 response buffers worth of rows, otherwise we need
+	 * 3 responses, of which the last one contains no rows. */
+	for (i = 0; i < ((2 * n_rows_buffer) - 1); i++) {
+		EXEC("INSERT INTO test(n) VALUES(123)");
+	}
+	EXEC("COMMIT");
+
+	f->request.db_id = 0;
+	f->request.sql = "SELECT n FROM test";
+	ENCODE(&f->request, query_sql);
+	HANDLE(QUERY_SQL);
+	ASSERT_CALLBACK(0, ROWS);
+
+	uint64__decode(f->cursor, &n);
+	munit_assert_int(n, ==, 1);
+	text__decode(f->cursor, &column);
+	munit_assert_string_equal(column, "n");
+
+	/* First response contains max amount of rows */
+	for (i = 0; i < n_rows_buffer; i++) {
+		DECODE_ROW(1, &value);
+		munit_assert_int(value.type, ==, SQLITE_INTEGER);
+		munit_assert_int(value.integer, ==, 123);
+	}
+
+	DECODE(&f->response, rows);
+	munit_assert_ullong(f->response.eof, ==, DQLITE_RESPONSE_ROWS_PART);
+
+	return MUNIT_OK;
+}
+
 /* Perform a query with parameters */
 TEST_CASE(query_sql, params, NULL)
 {
